import unittest

import numpy as np
import numpy.testing as npt

import sigpy as sp
from sigpy.mri import linop

if __name__ == "__main__":
    unittest.main()


def check_linop_adjoint(A, dtype=float, device=sp.cpu_device):
    device = sp.Device(device)
    x = sp.randn(A.ishape, dtype=dtype, device=device)
    y = sp.randn(A.oshape, dtype=dtype, device=device)

    xp = device.xp
    with device:
        lhs = xp.vdot(A * x, y)
        rhs = xp.vdot(x, A.H * y)

        xp.testing.assert_allclose(lhs, rhs, atol=1e-5, rtol=1e-5)


class TestLinop(unittest.TestCase):
    def test_sense_model(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=complex)
        mps = sp.randn(mps_shape, dtype=complex)

        A = linop.Sense(mps)

        check_linop_adjoint(A, dtype=complex)

        npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]), A * img)

    def test_sense_model_batch(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=complex)
        mps = sp.randn(mps_shape, dtype=complex)

        for coil_batch_size in [None, 1, 2, 3]:
            A = linop.Sense(mps, coil_batch_size=coil_batch_size)
            check_linop_adjoint(A, dtype=complex)
            npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]), A * img)

    def test_noncart_sense_model(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=complex)
        mps = sp.randn(mps_shape, dtype=complex)

        y, x = np.mgrid[:16, :16]
        coord = np.stack([np.ravel(y - 8), np.ravel(x - 8)], axis=1)
        coord = coord.astype(float)

        A = linop.Sense(mps, coord=coord)
        check_linop_adjoint(A, dtype=complex)
        npt.assert_allclose(
            sp.fft(img * mps, axes=[-1, -2]).ravel(),
            (A * img).ravel(),
            atol=0.1,
            rtol=0.1,
        )

    def test_sense_tseg_off_res_model(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=complex)
        mps = sp.randn(mps_shape, dtype=complex)

        y, x = np.mgrid[:16, :16]
        coord = np.stack([np.ravel(y - 8), np.ravel(x - 8)], axis=1)
        coord = coord.astype(float)

        d = np.sqrt(x * x + y * y)
        sigma, mu, a = 2, 0.25, 400
        b0 = a * np.exp(-((d - mu) ** 2 / (2.0 * sigma**2)))
        tseg = {"b0": b0, "dt": 4e-6, "lseg": 1, "n_bins": 10}

        F = sp.linop.NUFFT(mps_shape, coord)
        b, ct = sp.mri.util.tseg_off_res_b_ct(
            b0=b0, bins=10, lseg=1, dt=4e-6, T=coord.shape[0] * 4e-6
        )
        B1 = sp.linop.Multiply(F.oshape, b.T)
        Ct1 = sp.linop.Multiply(img_shape, ct.reshape(img_shape))
        S = sp.linop.Multiply(img_shape, mps)

        A = linop.Sense(mps, coord=coord, tseg=tseg)

        check_linop_adjoint(A, dtype=complex)
        npt.assert_allclose(B1 * F * S * Ct1 * img, A * img)

    def test_noncart_sense_model_batch(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=complex)
        mps = sp.randn(mps_shape, dtype=complex)

        y, x = np.mgrid[:16, :16]
        coord = np.stack([np.ravel(y - 8), np.ravel(x - 8)], axis=1)
        coord = coord.astype(float)

        for coil_batch_size in [None, 1, 2, 3]:
            A = linop.Sense(mps, coord=coord, coil_batch_size=coil_batch_size)
            check_linop_adjoint(A, dtype=complex)
            npt.assert_allclose(
                sp.fft(img * mps, axes=[-1, -2]).ravel(),
                (A * img).ravel(),
                atol=0.1,
                rtol=0.1,
            )

    if sp.config.mpi4py_enabled:

        def test_sense_model_with_comm(self):
            img_shape = [16, 16]
            mps_shape = [8, 16, 16]
            comm = sp.Communicator()

            img = sp.randn(img_shape, dtype=complex)
            mps = sp.randn(mps_shape, dtype=complex)
            comm.allreduce(img)
            comm.allreduce(mps)
            ksp = sp.fft(img * mps, axes=[-1, -2])

            A = linop.Sense(mps[comm.rank :: comm.size], comm=comm)

            npt.assert_allclose(
                A.H(ksp[comm.rank :: comm.size]),
                np.sum(sp.ifft(ksp, axes=[-1, -2]) * mps.conjugate(), 0),
            )



# -*- coding: utf-8 -*-
"""FFT and non-uniform FFT (NUFFT) functions.

"""
from math import ceil

import numpy as np

from sigpy import backend, interp, util

__all__ = [
    "fft",
    "ifft",
    "nufft",
    "nufft_adjoint",
    "estimate_shape",
    "toeplitz_psf",
]


def fft(input, oshape=None, axes=None, center=True, norm="ortho"):
    """FFT function that supports centering.

    Args:
        input (array): input array.
        oshape (None or array of ints): output shape.
        axes (None or array of ints): Axes over which to compute the FFT.
        norm (Nonr or ``"ortho"``): Keyword to specify the normalization mode.

    Returns:
        array: FFT result of dimension oshape.

    See Also:
        :func:`numpy.fft.fftn`

    """
    xp = backend.get_array_module(input)
    if not np.issubdtype(input.dtype, np.complexfloating):
        input = input.astype(np.complex64)

    if center:
        output = _fftc(input, oshape=oshape, axes=axes, norm=norm)
    else:
        output = xp.fft.fftn(input, s=oshape, axes=axes, norm=norm)

    if (
        np.issubdtype(input.dtype, np.complexfloating)
        and input.dtype != output.dtype
    ):
        output = output.astype(input.dtype, copy=False)

    return output


def ifft(input, oshape=None, axes=None, center=True, norm="ortho"):
    """IFFT function that supports centering.

    Args:
        input (array): input array.
        oshape (None or array of ints): output shape.
        axes (None or array of ints): Axes over which to compute
            the inverse FFT.
        norm (None or ``"ortho"``): Keyword to specify the normalization mode.

    Returns:
        array of dimension oshape.

    See Also:
        :func:`numpy.fft.ifftn`

    """
    xp = backend.get_array_module(input)
    if not np.issubdtype(input.dtype, np.complexfloating):
        input = input.astype(np.complex64)

    if center:
        output = _ifftc(input, oshape=oshape, axes=axes, norm=norm)
    else:
        output = xp.fft.ifftn(input, s=oshape, axes=axes, norm=norm)

    if (
        np.issubdtype(input.dtype, np.complexfloating)
        and input.dtype != output.dtype
    ):
        output = output.astype(input.dtype)

    return output


def nufft(input, coord, oversamp=1.25, width=4):
    """Non-uniform Fast Fourier Transform.

    Args:
        input (array): input signal domain array of shape
            (..., n_{ndim - 1}, ..., n_1, n_0),
            where ndim is specified by coord.shape[-1]. The nufft
            is applied on the last ndim axes, and looped over
            the remaining axes.
        coord (array): Fourier domain coordinate array of shape (..., ndim).
            ndim determines the number of dimensions to apply the nufft.
            coord[..., i] should be scaled to have its range between
            -n_i // 2, and n_i // 2.
        oversamp (float): oversampling factor.
        width (float): interpolation kernel full-width in terms of
            oversampled grid.

    Returns:
        array: Fourier domain data of shape
            input.shape[:-ndim] + coord.shape[:-1].

    References:
        Fessler, J. A., & Sutton, B. P. (2003).
        Nonuniform fast Fourier transforms using min-max interpolation
        IEEE Transactions on Signal Processing, 51(2), 560-574.
        Beatty, P. J., Nishimura, D. G., & Pauly, J. M. (2005).
        Rapid gridding reconstruction with a minimal oversampling ratio.
        IEEE transactions on medical imaging, 24(6), 799-808.

    """
    ndim = coord.shape[-1]
    beta = np.pi * (((width / oversamp) * (oversamp - 0.5)) ** 2 - 0.8) ** 0.5
    os_shape = _get_oversamp_shape(input.shape, ndim, oversamp)

    output = input.copy()

    # Apodize
    _apodize(output, ndim, oversamp, width, beta)

    # Zero-pad
    output /= util.prod(input.shape[-ndim:]) ** 0.5
    output = util.resize(output, os_shape)

    # FFT
    output = fft(output, axes=range(-ndim, 0), norm=None)

    # Interpolate
    coord = _scale_coord(coord, input.shape, oversamp)
    output = interp.interpolate(
        output, coord, kernel="kaiser_bessel", width=width, param=beta
    )
    output /= width**ndim

    return output


def estimate_shape(coord):
    """Estimate array shape from coordinates.

    Shape is estimated by the different between maximum and minimum of
    coordinates in each axis.

    Args:
        coord (array): Coordinates.
    """
    ndim = coord.shape[-1]
    with backend.get_device(coord):
        shape = [
            int(coord[..., i].max() - coord[..., i].min()) for i in range(ndim)
        ]

    return shape


def nufft_adjoint(input, coord, oshape=None, oversamp=1.25, width=4):
    """Adjoint non-uniform Fast Fourier Transform.

    Args:
        input (array): input Fourier domain array of shape
            (...) + coord.shape[:-1]. That is, the last dimensions
            of input must match the first dimensions of coord.
            The nufft_adjoint is applied on the last coord.ndim - 1 axes,
            and looped over the remaining axes.
        coord (array): Fourier domain coordinate array of shape (..., ndim).
            ndim determines the number of dimension to apply nufft adjoint.
            coord[..., i] should be scaled to have its range between
            -n_i // 2, and n_i // 2.
        oshape (tuple of ints): output shape of the form
            (..., n_{ndim - 1}, ..., n_1, n_0).
        oversamp (float): oversampling factor.
        width (float): interpolation kernel full-width in terms of
            oversampled grid.

    Returns:
        array: signal domain array with shape specified by oshape.

    See Also:
        :func:`sigpy.nufft.nufft`

    """
    ndim = coord.shape[-1]
    beta = np.pi * (((width / oversamp) * (oversamp - 0.5)) ** 2 - 0.8) ** 0.5
    if oshape is None:
        oshape = list(input.shape[: -coord.ndim + 1]) + estimate_shape(coord)
    else:
        oshape = list(oshape)

    os_shape = _get_oversamp_shape(oshape, ndim, oversamp)

    # Gridding
    coord = _scale_coord(coord, oshape, oversamp)
    output = interp.gridding(
        input, coord, os_shape, kernel="kaiser_bessel", width=width, param=beta
    )
    output /= width**ndim

    # IFFT
    output = ifft(output, axes=range(-ndim, 0), norm=None)

    # Crop
    output = util.resize(output, oshape)
    output *= util.prod(os_shape[-ndim:]) / util.prod(oshape[-ndim:]) ** 0.5

    # Apodize
    _apodize(output, ndim, oversamp, width, beta)

    return output


def toeplitz_psf(coord, shape, oversamp=1.25, width=4):
    """Toeplitz PSF for fast Normal non-uniform Fast Fourier Transform.

    While fast, this is more memory intensive.

    Args:
        coord (array): Fourier domain coordinate array of shape (..., ndim).
            ndim determines the number of dimension to apply nufft adjoint.
            coord[..., i] should be scaled to have its range between
            -n_i // 2, and n_i // 2.
        shape (tuple of ints): shape of the form
            (..., n_{ndim - 1}, ..., n_1, n_0).
            This is the shape of the input array of the forward nufft.
        oversamp (float): oversampling factor.
        width (float): interpolation kernel full-width in terms of
            oversampled grid.

    Returns:
        array: PSF to be used by the normal operator defined in
            `sigpy.linop.NUFFT`

    See Also:
        :func:`sigpy.linop.NUFFT`

    """
    xp = backend.get_array_module(coord)
    with backend.get_device(coord):
        ndim = coord.shape[-1]

        new_shape = _get_oversamp_shape(shape, ndim, 2)
        new_coord = _scale_coord(coord, new_shape, 2)

        idx = [slice(None)] * len(new_shape)
        for k in range(-1, -(ndim + 1), -1):
            idx[k] = new_shape[k] // 2

        d = xp.zeros(new_shape, dtype=xp.complex64)
        d[tuple(idx)] = 1

        psf = nufft(d, new_coord, oversamp, width)
        psf = nufft_adjoint(psf, new_coord, d.shape, oversamp, width)
        fft_axes = tuple(range(-1, -(ndim + 1), -1))
        psf = fft(psf, axes=fft_axes, norm=None) * (2**ndim)

        return psf


def _fftc(input, oshape=None, axes=None, norm="ortho"):
    ndim = input.ndim
    axes = util._normalize_axes(axes, ndim)
    xp = backend.get_array_module(input)

    if oshape is None:
        oshape = input.shape

    tmp = util.resize(input, oshape)
    tmp = xp.fft.ifftshift(tmp, axes=axes)
    tmp = xp.fft.fftn(tmp, axes=axes, norm=norm)
    output = xp.fft.fftshift(tmp, axes=axes)
    return output


def _ifftc(input, oshape=None, axes=None, norm="ortho"):
    ndim = input.ndim
    axes = util._normalize_axes(axes, ndim)
    xp = backend.get_array_module(input)

    if oshape is None:
        oshape = input.shape

    tmp = util.resize(input, oshape)
    tmp = xp.fft.ifftshift(tmp, axes=axes)
    tmp = xp.fft.ifftn(tmp, axes=axes, norm=norm)
    output = xp.fft.fftshift(tmp, axes=axes)
    return output


def _scale_coord(coord, shape, oversamp):
    ndim = coord.shape[-1]
    output = coord.copy()
    for i in range(-ndim, 0):
        scale = ceil(oversamp * shape[i]) / shape[i]
        shift = ceil(oversamp * shape[i]) // 2
        output[..., i] *= scale
        output[..., i] += shift

    return output


def _get_oversamp_shape(shape, ndim, oversamp):
    return list(shape)[:-ndim] + [ceil(oversamp * i) for i in shape[-ndim:]]


def _apodize(input, ndim, oversamp, width, beta):
    xp = backend.get_array_module(input)
    output = input
    for a in range(-ndim, 0):
        i = output.shape[a]
        os_i = ceil(oversamp * i)
        idx = xp.arange(i, dtype=output.dtype)

        # Calculate apodization
        apod = (
            beta**2 - (np.pi * width * (idx - i // 2) / os_i) ** 2
        ) ** 0.5
        apod /= xp.sinh(apod)
        output *= apod.reshape([i] + [1] * (-a - 1))

    return output



import unittest

import numpy as np
import numpy.testing as npt
from scipy.ndimage import gaussian_filter

import sigpy as sp
from sigpy.mri import linop, rf, sim

if __name__ == "__main__":
    unittest.main()


class TestPtx(unittest.TestCase):
    @staticmethod
    def problem_2d(dim):
        img_shape = [dim, dim]
        sens_shape = [8, dim, dim]

        # target - slightly blurred circle
        x, y = np.ogrid[
            -img_shape[0] / 2 : img_shape[0] - img_shape[0] / 2,
            -img_shape[1] / 2 : img_shape[1] - img_shape[1] / 2,
        ]
        circle = x * x + y * y <= int(img_shape[0] / 6) ** 2
        target = np.zeros(img_shape)
        target[circle] = 1
        target = gaussian_filter(target, 1)
        target = target.astype(np.complex64)

        sens = sim.birdcage_maps(sens_shape)

        return target, sens

    @staticmethod
    def problem_3d(dim, Nz):
        Nc = 8
        img_shape = [dim, dim, Nz]
        sens_shape = [Nc, dim, dim, Nz]

        # target - slightly blurred circle
        x, y, z = np.ogrid[
            -img_shape[0] / 2 : img_shape[0] - img_shape[0] / 2,
            -img_shape[1] / 2 : img_shape[1] - img_shape[1] / 2,
            -img_shape[2] / 2 : img_shape[2] - img_shape[2] / 2,
        ]
        circle = x * x + y * y + z * z <= int(img_shape[0] / 5) ** 2
        target = np.zeros(img_shape)
        target[circle] = 1
        target = gaussian_filter(target, 1)
        target = target.astype(np.complex64)
        sens = sp.mri.sim.birdcage_maps(sens_shape)

        return target, sens

    def test_stspa_radial(self):
        target, sens = self.problem_2d(8)

        # makes dim*dim*2 trajectory
        traj = sp.mri.radial(
            (sens.shape[1], sens.shape[1], 2),
            target.shape,
            golden=True,
            dtype=np.float32,
        )
        # reshape to be Nt*2 trajectory
        traj = np.reshape(traj, [traj.shape[0] * traj.shape[1], 2])

        A = linop.Sense(sens, coord=traj, weights=None, ishape=target.shape).H

        pulses = rf.stspa(
            target,
            sens,
            traj,
            dt=4e-6,
            alpha=1,
            b0=None,
            st=None,
            explicit=False,
            max_iter=100,
            tol=1e-4,
        )

        npt.assert_array_almost_equal(A * pulses, target, 1e-3)

    def test_stspa_spiral(self):
        target, sens = self.problem_2d(8)

        fov = 0.55
        gts = 6.4e-6
        gslew = 190
        gamp = 40
        R = 1
        dx = 0.025  # in m
        # construct a trajectory
        g, k, t, s = rf.spiral_arch(fov / R, dx, gts, gslew, gamp)

        A = linop.Sense(sens, coord=k, ishape=target.shape).H

        pulses = rf.stspa(
            target,
            sens,
            k,
            dt=4e-6,
            alpha=1,
            b0=None,
            st=None,
            explicit=False,
            max_iter=100,
            tol=1e-4,
        )

        npt.assert_array_almost_equal(A * pulses, target, 1e-3)

    def test_stspa_2d_explicit(self):
        target, sens = self.problem_2d(8)
        dim = target.shape[0]
        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        A = rf.PtxSpatialExplicit(
            sens, k1, dt=4e-6, img_shape=target.shape, b0=None
        )
        pulses = sp.mri.rf.stspa(
            target,
            sens,
            st=None,
            coord=k1,
            dt=4e-6,
            max_iter=100,
            alpha=10,
            tol=1e-4,
            phase_update_interval=200,
            explicit=True,
        )

        npt.assert_array_almost_equal(A * pulses, target, 1e-3)

    def test_stspa_3d_explicit(self):
        nz = 4
        target, sens = self.problem_3d(3, nz)
        dim = target.shape[0]

        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        k1 = rf.stack_of(k1, nz, 0.1)
        A = rf.linop.PtxSpatialExplicit(
            sens, k1, dt=4e-6, img_shape=target.shape, b0=None
        )

        pulses = sp.mri.rf.stspa(
            target,
            sens,
            st=None,
            coord=k1,
            dt=4e-6,
            max_iter=30,
            alpha=10,
            tol=1e-3,
            phase_update_interval=200,
            explicit=True,
        )

        npt.assert_array_almost_equal(A * pulses, target, 1e-3)

    def test_stspa_3d_nonexplicit(self):
        nz = 3
        target, sens = self.problem_3d(3, nz)
        dim = target.shape[0]

        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        k1 = rf.stack_of(k1, nz, 0.1)
        A = sp.mri.linop.Sense(
            sens, k1, weights=None, tseg=None, ishape=target.shape
        ).H

        pulses = sp.mri.rf.stspa(
            target,
            sens,
            st=None,
            coord=k1,
            dt=4e-6,
            max_iter=30,
            alpha=10,
            tol=1e-3,
            phase_update_interval=200,
            explicit=False,
        )

        npt.assert_array_almost_equal(A * pulses, target, 1e-3)

    def test_spokes(self):
        # spokes problem definition:
        dim = 20  # size of the b1 matrix loaded
        n_spokes = 5
        fov = 20  # cm
        dx_max = 2  # cm
        gts = 4e-6
        sl_thick = 5  # slice thickness, mm
        tbw = 4
        dgdtmax = 18000  # g/cm/s
        gmax = 2  # g/cm

        _, sens = self.problem_2d(dim)
        roi = np.zeros((dim, dim))
        radius = dim // 2
        cx, cy = dim // 2, dim // 2
        y, x = np.ogrid[-radius:radius, -radius:radius]
        index = x**2 + y**2 <= radius**2
        roi[cy - radius : cy + radius, cx - radius : cx + radius][index] = 1
        sens = sens * roi

        [pulses, g] = rf.stspk(
            roi,
            sens,
            n_spokes,
            fov,
            dx_max,
            gts,
            sl_thick,
            tbw,
            dgdtmax,
            gmax,
            alpha=1,
        )

        # should give the number of pulses corresponding to number of TX ch
        npt.assert_equal(np.shape(pulses)[0], np.shape(sens)[0])
        # should hit the max gradient constraint
        npt.assert_almost_equal(gmax, np.max(g), decimal=3)



# -*- coding: utf-8 -*-
"""Functions and classes for getting and setting computing devices.

"""
import numpy as np

from sigpy import config

if config.cupy_enabled:
    import cupy as cp

if config.mpi4py_enabled:
    from mpi4py import MPI

    if config.nccl_enabled:
        from cupy.cuda import nccl


__all__ = [
    "Device",
    "get_device",
    "get_array_module",
    "cpu_device",
    "to_device",
    "copyto",
    "Communicator",
]


class Device(object):
    """Device class.

    This class extends cupy.Device, with id > 0 representing the id_th GPU,
    and id = -1 representing CPU. cupy must be installed to use GPUs.

    The array module for the corresponding device can be obtained via .xp.
    Similar to cupy.Device, the Device object can be used as a context:

        >>> device = Device(2)
        >>> xp = device.xp  # xp is cupy.
        >>> with device:
        >>>     x = xp.array([1, 2, 3])
        >>>     x += 1

    Args:
        id_or_device (int or Device or cupy.cuda.Device): id > 0 represents
            the corresponding GPUs, and id = -1 represents CPU.

    Attributes:
        id (int): id = -1 represents CPU,
            and others represents the id_th GPUs.

    """

    def __init__(self, id_or_device):
        if isinstance(id_or_device, int):
            id = id_or_device
        elif isinstance(id_or_device, Device):
            id = id_or_device.id
        elif config.cupy_enabled and isinstance(id_or_device, cp.cuda.Device):
            id = id_or_device.id
        else:
            raise ValueError(
                "Accepts int, Device or cupy.cuda.Device, got {}".format(
                    id_or_device
                )
            )

        if id != -1:
            if config.cupy_enabled:
                self.cpdevice = cp.cuda.Device(id)
            else:
                raise ValueError(
                    "cupy not installed, but set device {}.".format(id)
                )

        self.id = id

    @property
    def xp(self):
        """module: numpy or cupy module for the device."""
        if self.id == -1:
            return np

        return cp

    def use(self):
        """Use computing device.

        All operations after use() will use the device.
        """
        if self.id > 0:
            self.cpdevice.use()

    def __int__(self):
        return self.id

    def __eq__(self, other):
        if isinstance(other, int):
            return self.id == other
        elif isinstance(other, Device):
            return self.id == other.id
        elif config.cupy_enabled and isinstance(other, cp.cuda.Device):
            return self.id == other.id
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __enter__(self):
        if self.id == -1:
            return None

        return self.cpdevice.__enter__()

    def __exit__(self, *args):
        if self.id == -1:
            pass
        else:
            self.cpdevice.__exit__()

    def __repr__(self):
        if self.id == -1:
            return "<CPU Device>"

        return self.cpdevice.__repr__()


cpu_device = Device(-1)


def get_array_module(array):
    """Gets an appropriate module from :mod:`numpy` or :mod:`cupy`.

    This is almost equivalent to :func:`cupy.get_array_module`. The differences
    are that this function can be used even if cupy is not available.

    Args:
        array: Input array.

    Returns:
        module: :mod:`cupy` or :mod:`numpy` is returned based on input.
    """
    if config.cupy_enabled:
        return cp.get_array_module(array)
    else:
        return np


def get_device(array):
    """Get Device from input array.

    Args:
        array (array): Array.

    Returns:
        Device.

    """
    if get_array_module(array) == np:
        return cpu_device
    else:
        return Device(array.device)


def to_device(input, device=cpu_device):
    """Move input to device. Does not copy if same device.

    Args:
        input (array): Input.
        device (int or Device or cupy.Device): Output device.

    Returns:
        array: Output array placed in device.
    """
    idevice = get_device(input)
    odevice = Device(device)

    if idevice == odevice:
        return input

    if odevice == cpu_device:
        with idevice:
            return input.get()
    else:
        with odevice:
            return cp.asarray(input)


def copyto(output, input):
    """Copy from input to output. Input/output can be in different device.

    Args:
        input (array): Input.
        output (array): Output.

    """
    idevice = get_device(input)
    odevice = get_device(output)
    if idevice == cpu_device and odevice != cpu_device:
        with odevice:
            output.set(input)
    elif idevice != cpu_device and odevice == cpu_device:
        with idevice:
            np.copyto(output, input.get())
    else:
        idevice.xp.copyto(output, input)


class Communicator(object):
    """Communicator for distributed computing using MPI.

    When NCCL is not installed, arrays are moved to CPU,
    then communicated through MPI, and moved back
    to original device.
    When mpi4py is not installed, the communicator errors.

    """

    def __init__(self):
        if config.mpi4py_enabled:
            self.mpi_comm = MPI.COMM_WORLD
            self.size = self.mpi_comm.Get_size()
            self.rank = self.mpi_comm.Get_rank()
        else:
            self.size = 1
            self.rank = 0

        # Keep nccl comms for reuse
        if config.nccl_enabled:
            self.nccl_comms = {}

    def allreduce(self, input):
        """All reduce operation in-place.

        Sums input across all nodes and broadcast back to each node.

        Args:
            input (array): input array.

        """
        if self.size > 1:
            if config.nccl_enabled:
                device = get_device(input)
                devices = self.mpi_comm.allgather(device.id)
                if all([d >= 0 for d in devices]):
                    nccl_comm = self._get_nccl_comm(device, devices)
                    nccl_dtype, nccl_size = self._get_nccl_dtype_size(input)
                    with device:
                        nccl_comm.allReduce(
                            input.data.ptr,
                            input.data.ptr,
                            nccl_size,
                            nccl_dtype,
                            nccl.NCCL_SUM,
                            cp.cuda.Stream.null.ptr,
                        )
                        return

            cpu_input = to_device(input, cpu_device)
            self.mpi_comm.Allreduce(MPI.IN_PLACE, cpu_input)
            copyto(input, cpu_input)

    def reduce(self, input, root=0):
        """Reduce operation in-place.

        Sums input across all nodes in root node.

        Args:
            input (array): input array.
            root (int): root node rank.

        """
        if self.size > 1:
            cpu_input = to_device(input, cpu_device)
            if self.rank == root:
                self.mpi_comm.Reduce(MPI.IN_PLACE, cpu_input, root=root)
                copyto(input, cpu_input)
            else:
                self.mpi_comm.Reduce(cpu_input, None, root=root)

    def bcast(self, input, root=0, datatype=None):
        """Broadcast from root to other nodes.

        Args:
            input (array): input array.
            root (int): root node rank.
            datatype (int): MPI datatype for broadcasting.

        """
        if config.mpi4py_enabled:
            datatype = MPI.COMPLEX

        if self.size > 1:
            cpu_input = to_device(input, cpu_device)
            self.mpi_comm.Bcast((cpu_input, datatype), root=root)
            copyto(input, cpu_input)

    def gatherv(self, input, root=0):
        """Gather with variable sizes operation.

        Gather inputs across all nodes to the root node,
        and vectorizes them.

        Args:
            input (array): input array.
            root (int): root node rank.

        Returns:
            array or None: vectorized array if rank==root else None.

        """
        if self.size > 1:
            cpu_input = to_device(input, cpu_device)

            sizes = self.mpi_comm.gather(input.size, root=root)
            if self.rank == root:
                cpu_output = np.empty(sum(sizes), dtype=input.dtype)
                self.mpi_comm.Gatherv(
                    cpu_input, [cpu_output, sizes], root=root
                )
                return to_device(cpu_output, get_device(input))
            else:
                self.mpi_comm.Gatherv(cpu_input, [None, sizes], root=root)
        else:
            return input

    def _get_nccl_comm(self, device, devices):
        if str(devices) in self.nccl_comms:
            return self.nccl_comms[str(devices)]

        if self.rank == 0:
            nccl_comm_id = nccl.get_unique_id()
        else:
            nccl_comm_id = None

        nccl_comm_id = self.mpi_comm.bcast(nccl_comm_id)

        with device:
            nccl_comm = nccl.NcclCommunicator(
                self.size, nccl_comm_id, self.rank
            )
            self.nccl_comms[str(devices)] = nccl_comm

        return nccl_comm

    def _get_nccl_dtype_size(self, input):
        if input.dtype == np.float32:
            nccl_dtype = nccl.NCCL_FLOAT32
            nccl_size = input.size
        elif input.dtype == np.float64:
            nccl_dtype = nccl.NCCL_FLOAT64
            nccl_size = input.size
        elif input.dtype == np.complex64:
            nccl_dtype = nccl.NCCL_FLOAT32
            nccl_size = input.size * 2
        elif input.dtype == np.complex128:
            nccl_dtype = nccl.NCCL_FLOAT64
            nccl_size = input.size * 2
        else:
            raise ValueError(
                "dtype not supported, got {dtype}.".format(dtype=input.dtype)
            )

        return nccl_dtype, nccl_size



import unittest

import numpy as np
import numpy.testing as npt

from sigpy.learn import app

if __name__ == "__main__":
    unittest.main()


class TestApp(unittest.TestCase):
    def test_ConvSparseDecom(self):
        lamda = 1e-9
        L = np.array([[1, 1], [1, -1]], dtype=float) / 2**0.5
        y = np.array([[1, 1]], dtype=float) / 2**0.5

        R = app.ConvSparseDecom(y, L, lamda=lamda).run()

        npt.assert_allclose(R, [[[1], [0]]])

    def test_ConvSparseCoefficients(self):
        lamda = 1e-10
        L = np.array([[1, 1], [1, -1]], dtype=float) / 2**0.5
        y = np.array([[1, 1]], dtype=float) / 2**0.5

        R_j = app.ConvSparseCoefficients(y, L, lamda=lamda)

        npt.assert_allclose(R_j[:], [[[1], [0]]])
        npt.assert_allclose(R_j[0, :], [[1], [0]])
        npt.assert_allclose(R_j[:, 0], [[1]])
        npt.assert_allclose(R_j[:, :, 0], [[1, 0]])

    def test_ConvSparseCoding(self):
        num_atoms = 1
        filt_width = 2
        batch_size = 1
        y = np.array([[1, 1]], dtype=float) / 2**0.5
        lamda = 1e-10
        alpha = 1

        L, _ = app.ConvSparseCoding(
            y,
            num_atoms,
            filt_width,
            batch_size,
            alpha=alpha,
            lamda=lamda,
            max_iter=100,
        ).run()

        npt.assert_allclose(
            np.abs(L), [[1 / 2**0.5, 1 / 2**0.5]], atol=0.1, rtol=0.1
        )

    def test_LinearRegression(self):
        n = 2
        k = 5
        m = 4
        batch_size = n

        X = np.random.randn(n, k)
        y = np.random.randn(n, m)

        alpha = 1 / np.linalg.svd(X, compute_uv=False)[0] ** 2
        mat = app.LinearRegression(X, y, batch_size, alpha, max_iter=300).run()
        mat_lstsq = np.linalg.lstsq(X, y, rcond=-1)[0]

        npt.assert_allclose(mat, mat_lstsq, atol=1e-2, rtol=1e-2)



import unittest

import numpy as np
import numpy.testing as npt

from sigpy.mri import rf

if __name__ == "__main__":
    unittest.main()


class TestAdiabatic(unittest.TestCase):
    def test_bir4(self):
        # test an excitation bir4 pulse
        n = 1176
        dt = 4e-6
        dw0 = 100 * np.pi / dt / n
        beta = 10
        kappa = np.arctan(20)
        flip = np.pi / 4
        [am_bir, om_bir] = rf.adiabatic.bir4(n, beta, kappa, flip, dw0)

        # check relatively homogeneous over range of B1 values
        b1 = np.arange(0.2, 0.8, 0.1)
        b1 = np.reshape(b1, (np.size(b1), 1))
        a = np.zeros(np.shape(b1), dtype="complex")
        b = np.zeros(np.shape(b1), dtype="complex")

        for ii in range(0, np.size(b1)):
            [a[ii], b[ii]] = rf.sim.abrm_nd(
                2 * np.pi * dt * 4258 * b1[ii] * am_bir,
                np.ones(1),
                dt * np.reshape(om_bir, (np.size(om_bir), 1)),
            )

        mxy = 2 * np.multiply(np.conj(a), b)

        test = np.ones(mxy.shape) * 0.7  # magnetization value we expect

        npt.assert_array_almost_equal(np.abs(mxy), test, 2)

    def test_hyp_ex(self):
        # test an inversion adiabatic hyp pulse
        n = 512
        beta = 800
        mu = 4.9
        dur = 0.012
        [am_sech, om_sech] = rf.adiabatic.hypsec(n, beta, mu, dur)

        # check relatively homogeneous over range of B1 values
        b1 = np.arange(0.2, 0.8, 0.1)
        b1 = np.reshape(b1, (np.size(b1), 1))

        a = np.zeros(np.shape(b1), dtype="complex")
        b = np.zeros(np.shape(b1), dtype="complex")
        for ii in range(0, np.size(b1)):
            [a[ii], b[ii]] = rf.sim.abrm_nd(
                2 * np.pi * (dur / n) * 4258 * b1[ii] * am_sech,
                np.ones(1),
                dur / n * np.reshape(om_sech, (np.size(om_sech), 1)),
            )
        mz = 1 - 2 * np.abs(b) ** 2

        test = np.ones(mz.shape) * -1  # magnetization value we expect

        npt.assert_array_almost_equal(mz, test, 2)

    def test_goia_wurst(self):
        # test a goia-wurst adiabatic pulse
        n = 512
        dur = 3.5e-3
        f = 0.9
        n_b1 = 16
        m_grad = 4
        [_, om_goia, g_goia] = rf.adiabatic.goia_wurst(n, dur, f, n_b1, m_grad)

        # test midpoint of goia pulse. Expect 1-f g, 0.1 fm
        npt.assert_almost_equal(g_goia[int(len(g_goia) / 2)], 1 - f, 2)
        npt.assert_almost_equal(g_goia[int(len(om_goia) / 2)], 0.1, 2)



import unittest

from sigpy import version

if __name__ == "__main__":
    unittest.main()


class TestVersion(unittest.TestCase):
    def test_version(self):
        assert version.__version__ == "0.1.25"



import unittest

import numpy as np
import numpy.testing as npt

from sigpy.mri import samp

if __name__ == "__main__":
    unittest.main()


class TestPoisson(unittest.TestCase):
    """Test poisson undersampling defined in `sigpy.mri.samp.poisson`."""

    def test_numpy_random_state(self):
        """Verify that random state is unchanged when seed is specified."""
        np.random.seed(0)
        expected_state = np.random.get_state()

        _ = samp.poisson((120, 120), accel=6, seed=80)

        state = np.random.get_state()
        assert (expected_state[1] == state[1]).all()

    def test_reproducibility(self):
        """Verify that poisson is reproducible."""
        np.random.seed(45)
        mask1 = samp.poisson((120, 120), accel=6, seed=80)

        # Changing internal numpy state should not affect mask.
        np.random.seed(20)
        mask2 = samp.poisson((120, 120), accel=6, seed=80)

        npt.assert_allclose(mask2, mask1)

    def test_poisson_accel(self):
        """Verify that poisson generates the correct acceleration."""
        for x in [60, 120]:
            for y in [60, 120]:
                for tol in [0.1, 0.2]:
                    for accel in [4, 5, 6, 7, 8]:
                        mask = samp.poisson(
                            (x, y), accel=accel, seed=80, tol=tol
                        )
                        assert abs(mask.size / np.sum(mask) - accel) < tol



"""The core module contains functions and classes for signal processing.

SigPy provides simple interfaces to commonly used signal processing functions,
including convolution, FFT, NUFFT, wavelet transform, and thresholdings.
All functions, except wavelet transform, can run on both CPU and GPU.

These functions are wrapped into higher level classes (Linop and Prox)
that can be used in conjuction with Alg to form an App.

"""
from sigpy import (
    alg,
    app,
    backend,
    block,
    config,
    conv,
    fourier,
    interp,
    linop,
    prox,
    pytorch,
    sim,
    thresh,
    util,
    wavelet,
)
from sigpy.backend import *  # noqa
from sigpy.block import *  # noqa
from sigpy.conv import *  # noqa
from sigpy.fourier import *  # noqa
from sigpy.interp import *  # noqa
from sigpy.pytorch import *  # noqa
from sigpy.sim import *  # noqa
from sigpy.thresh import *  # noqa
from sigpy.util import *  # noqa
from sigpy.wavelet import *  # noqa

from .version import __version__  # noqa

__all__ = ["alg", "app", "config", "linop", "prox"]
__all__.extend(backend.__all__)
__all__.extend(block.__all__)
__all__.extend(conv.__all__)
__all__.extend(interp.__all__)
__all__.extend(fourier.__all__)
__all__.extend(pytorch.__all__)
__all__.extend(sim.__all__)
__all__.extend(thresh.__all__)
__all__.extend(util.__all__)
__all__.extend(wavelet.__all__)



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import backend, config, conv, util

if __name__ == "__main__":
    unittest.main()

dtypes = [np.float32, np.float64, np.complex64, np.complex128]


class TestConv(unittest.TestCase):
    def test_convolve_valid(self):
        mode = "valid"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for D in [1, 2, 3]:
            for device in devices:
                xp = device.xp
                with device:
                    for dtype in dtypes:
                        with self.subTest(D=D, dtype=dtype, device=device):
                            data = util.dirac(
                                [3] + [1] * (D - 1), device=device, dtype=dtype
                            )
                            filt = xp.ones([3] + [1] * (D - 1), dtype=dtype)
                            output = backend.to_device(
                                conv.convolve(data, filt, mode=mode)
                            )
                            npt.assert_allclose(
                                output, np.ones([1] * D), atol=1e-5
                            )

                            data = util.dirac(
                                [3] + [1] * (D - 1), device=device, dtype=dtype
                            )
                            filt = xp.ones([2] + [1] * (D - 1), dtype=dtype)
                            output = backend.to_device(
                                conv.convolve(data, filt, mode=mode)
                            )
                            npt.assert_allclose(
                                output, np.ones([2] + [1] * (D - 1)), atol=1e-5
                            )

                            data = util.dirac(
                                [1, 3] + [1] * (D - 1),
                                device=device,
                                dtype=dtype,
                            )
                            filt = xp.ones(
                                [2, 1, 3] + [1] * (D - 1), dtype=dtype
                            )
                            output = backend.to_device(
                                conv.convolve(
                                    data, filt, mode=mode, multi_channel=True
                                ),
                                backend.cpu_device,
                            )
                            npt.assert_allclose(
                                output,
                                np.ones([2, 1] + [1] * (D - 1)),
                                atol=1e-5,
                            )

                            data = util.dirac(
                                [1, 3] + [1] * (D - 1),
                                device=device,
                                dtype=dtype,
                            )
                            filt = xp.ones(
                                [2, 1, 3] + [1] * (D - 1), dtype=dtype
                            )
                            strides = [2] + [1] * (D - 1)
                            output = backend.to_device(
                                conv.convolve(
                                    data,
                                    filt,
                                    mode=mode,
                                    strides=strides,
                                    multi_channel=True,
                                ),
                                backend.cpu_device,
                            )
                            npt.assert_allclose(
                                output,
                                np.ones([2, 1] + [1] * (D - 1)),
                                atol=1e-5,
                            )

    def test_convolve_full(self):
        mode = "full"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for D in [1, 2, 3]:
            for device in devices:
                xp = device.xp
                with device:
                    for dtype in dtypes:
                        with self.subTest(D=D, dtype=dtype, device=device):
                            data = util.dirac(
                                [3] + [1] * (D - 1), device=device, dtype=dtype
                            )
                            filt = xp.ones([3] + [1] * (D - 1), dtype=dtype)
                            output = backend.to_device(
                                conv.convolve(data, filt, mode=mode)
                            )
                            npt.assert_allclose(
                                output,
                                np.array([0, 1, 1, 1, 0]).reshape(
                                    [5] + [1] * (D - 1)
                                ),
                                atol=1e-5,
                            )

                            data = util.dirac(
                                [3] + [1] * (D - 1), device=device, dtype=dtype
                            )
                            filt = xp.ones([2] + [1] * (D - 1), dtype=dtype)
                            output = backend.to_device(
                                conv.convolve(data, filt, mode=mode)
                            )
                            npt.assert_allclose(
                                output,
                                np.array([0, 1, 1, 0]).reshape(
                                    [4] + [1] * (D - 1)
                                ),
                                atol=1e-5,
                            )

                            data = util.dirac(
                                [1, 3] + [1] * (D - 1),
                                device=device,
                                dtype=dtype,
                            )
                            filt = xp.ones(
                                [2, 1, 3] + [1] * (D - 1), dtype=dtype
                            )
                            output = backend.to_device(
                                conv.convolve(
                                    data, filt, mode=mode, multi_channel=True
                                ),
                                backend.cpu_device,
                            )
                            npt.assert_allclose(
                                output,
                                np.array(
                                    [[0, 1, 1, 1, 0], [0, 1, 1, 1, 0]]
                                ).reshape([2, 5] + [1] * (D - 1)),
                                atol=1e-5,
                            )

                            data = util.dirac(
                                [1, 3] + [1] * (D - 1),
                                device=device,
                                dtype=dtype,
                            )
                            filt = xp.ones(
                                [2, 1, 3] + [1] * (D - 1), dtype=dtype
                            )
                            strides = [2] + [1] * (D - 1)
                            output = backend.to_device(
                                conv.convolve(
                                    data,
                                    filt,
                                    mode=mode,
                                    strides=strides,
                                    multi_channel=True,
                                )
                            )
                            npt.assert_allclose(
                                output,
                                np.array([[0, 1, 0], [0, 1, 0]]).reshape(
                                    [2, 3] + [1] * (D - 1)
                                ),
                                atol=1e-5,
                            )

    def test_convolve_data_adjoint_valid(self):
        mode = "valid"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        output = xp.ones([1, 1], dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(data, [[1, 1, 1]], atol=1e-5)

                        output = xp.ones([1, 2], dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(data, [[1, 2, 1]], atol=1e-5)

                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output,
                                filt,
                                data_shape,
                                mode=mode,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(data, [[[2, 2, 2]]], atol=1e-5)

                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 4]
                        strides = [1, 2]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output,
                                filt,
                                data_shape,
                                mode=mode,
                                strides=strides,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(data, [[[2, 2, 2, 0]]], atol=1e-5)

    def test_convolve_data_adjoint_full(self):
        mode = "full"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        output = xp.ones([1, 5], dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(data, [[3, 3, 3]], atol=1e-5)

                        output = xp.ones([1, 4], dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(data, [[2, 2, 2]], atol=1e-5)

                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output,
                                filt,
                                data_shape,
                                mode=mode,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(data, [[[6, 6, 6]]], atol=1e-5)

                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 8]
                        strides = [1, 2]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output,
                                filt,
                                data_shape,
                                mode=mode,
                                strides=strides,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(
                            data, [[[4, 2, 4, 2, 4, 2, 4, 2]]], atol=1e-5
                        )

    def test_convolve_filter_adjoint_valid(self):
        mode = "valid"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 1], dtype=dtype)
                        filt_shape = [1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(filt, [[1, 1, 1]], atol=1e-5)

                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 2], dtype=dtype)
                        filt_shape = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(filt, [[2, 2]], atol=1e-5)

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output,
                                data,
                                filt_shape,
                                mode=mode,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(
                            filt, [[[[1, 1, 1]]], [[[1, 1, 1]]]], atol=1e-5
                        )

                        data = xp.ones([1, 1, 4], dtype=dtype)
                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        strides = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output,
                                data,
                                filt_shape,
                                mode=mode,
                                strides=strides,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(
                            filt, [[[[1, 1, 1]]], [[[1, 1, 1]]]], atol=1e-5
                        )

    def test_convolve_filter_adjoint_full(self):
        mode = "full"
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 5], dtype=dtype)
                        filt_shape = [1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(filt, [[3, 3, 3]], atol=1e-5)

                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 4], dtype=dtype)
                        filt_shape = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode
                            )
                        )
                        npt.assert_allclose(filt, [[3, 3]], atol=1e-5)

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output,
                                data,
                                filt_shape,
                                mode=mode,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(
                            filt, [[[[3, 3, 3]]], [[[3, 3, 3]]]], atol=1e-5
                        )

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 3], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        strides = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output,
                                data,
                                filt_shape,
                                mode=mode,
                                strides=strides,
                                multi_channel=True,
                            ),
                            backend.cpu_device,
                        )
                        npt.assert_allclose(
                            filt, [[[[2, 1, 2]]], [[[2, 1, 2]]]], atol=1e-5
                        )



# -*- coding: utf-8 -*-
"""MRI applications.
"""
import numpy as np

import sigpy as sp
from sigpy.mri import linop

__all__ = [
    "SenseRecon",
    "L1WaveletRecon",
    "TotalVariationRecon",
    "JsenseRecon",
    "EspiritCalib",
]


def _estimate_weights(y, weights, coord):
    if weights is None and coord is None:
        with sp.get_device(y):
            weights = (sp.rss(y, axes=(0,)) > 0).astype(y.dtype)

    return weights


class SenseRecon(sp.app.LinearLeastSquares):
    r"""SENSE Reconstruction.

    Considers the problem

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 +
        \frac{\lambda}{2} \| x \|_2^2

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        tseg (None or Dictionary): parameters for time-segmented off-resonance
            correction. Parameters are 'b0' (array), 'dt' (float),
            'lseg' (int), and 'n_bins' (int). Lseg is the number of
            time segments used, and n_bins is the number of histogram bins.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
            Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Pruessmann, K. P., Weiger, M., Scheidegger, M. B., & Boesiger, P.
        (1999).
        SENSE: sensitivity encoding for fast MRI.
        Magnetic resonance in medicine, 42(5), 952-962.

        Pruessmann, K. P., Weiger, M., Bornert, P., & Boesiger, P. (2001).
        Advances in sensitivity encoding with arbitrary k-space trajectories.
        Magnetic resonance in medicine, 46(4), 638-651.

    """

    def __init__(
        self,
        y,
        mps,
        lamda=0,
        weights=None,
        tseg=None,
        coord=None,
        device=sp.cpu_device,
        coil_batch_size=None,
        comm=None,
        show_pbar=True,
        transp_nufft=False,
        **kwargs
    ):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(
            mps,
            coord=coord,
            weights=weights,
            tseg=tseg,
            coil_batch_size=coil_batch_size,
            comm=comm,
            transp_nufft=transp_nufft,
        )

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, lamda=lamda, show_pbar=show_pbar, **kwargs)


class L1WaveletRecon(sp.app.LinearLeastSquares):
    r"""L1 Wavelet regularized reconstruction.

    Considers the problem

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 + \lambda \| W x \|_1

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, W is the wavelet operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        wave_name (str): wavelet name.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Lustig, M., Donoho, D., & Pauly, J. M. (2007).
        Sparse MRI: The application of compressed sensing for rapid MR imaging.
        Magnetic Resonance in Medicine, 58(6), 1082-1195.

    """

    def __init__(
        self,
        y,
        mps,
        lamda,
        weights=None,
        coord=None,
        wave_name="db4",
        device=sp.cpu_device,
        coil_batch_size=None,
        comm=None,
        show_pbar=True,
        transp_nufft=False,
        **kwargs
    ):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(
            mps,
            coord=coord,
            weights=weights,
            comm=comm,
            coil_batch_size=coil_batch_size,
            transp_nufft=transp_nufft,
        )
        img_shape = mps.shape[1:]
        W = sp.linop.Wavelet(img_shape, wave_name=wave_name)
        proxg = sp.prox.UnitaryTransform(sp.prox.L1Reg(W.oshape, lamda), W)

        def g(input):
            device = sp.get_device(input)
            xp = device.xp
            with device:
                return lamda * xp.sum(xp.abs(W(input))).item()

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg=proxg, g=g, show_pbar=show_pbar, **kwargs)


class TotalVariationRecon(sp.app.LinearLeastSquares):
    r"""Total variation regularized reconstruction.

    Considers the problem:

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 + \lambda \| G x \|_1

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, G is the gradient operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Block, K. T., Uecker, M., & Frahm, J. (2007).
        Undersampled radial MRI with multiple coils.
        Iterative image reconstruction using a total variation constraint.
        Magnetic Resonance in Medicine, 57(6), 1086-1098.

    """

    def __init__(
        self,
        y,
        mps,
        lamda,
        weights=None,
        coord=None,
        device=sp.cpu_device,
        coil_batch_size=None,
        comm=None,
        show_pbar=True,
        transp_nufft=False,
        **kwargs
    ):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(
            mps,
            coord=coord,
            weights=weights,
            comm=comm,
            coil_batch_size=coil_batch_size,
            transp_nufft=transp_nufft,
        )

        G = sp.linop.FiniteDifference(A.ishape)
        proxg = sp.prox.L1Reg(G.oshape, lamda)

        def g(x):
            device = sp.get_device(x)
            xp = device.xp
            with device:
                return lamda * xp.sum(xp.abs(x)).item()

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(
            A, y, proxg=proxg, g=g, G=G, show_pbar=show_pbar, **kwargs
        )


class JsenseRecon(sp.app.App):
    r"""JSENSE/NLINV reconstruction.

    Considers the problem

    .. math::
        \min_{l, r} \frac{1}{2} \| l \ast r - y \|_2^2 +
        \frac{\lambda}{2} (\| l \|_2^2 + \| r \|_2^2)

    where :math:`\ast` is the convolution operator.

    This formulation with regularization corresponds to the version
    described in the NLINV paper. Without regularization (which is the
    default) this corresponds to the version from the JSENSE paper but using a
    truncated Fourier representation of the coils (as in NLINV) instead
    of polynomials.

    Args:
        y (array): k-space measurements.
        mps_ker_width (int): sensitivity maps kernel width.
        ksp_calib_width (int): k-space calibration width.
        lamda (float): regularization parameter.
        device (Device): device to perform reconstruction.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        img_shape (None or list): Image shape.
        grd_shape (None or list): Shape of grid.
        max_iter (int): Maximum number of iterations.
        max_inner_iter (int): Maximum number of inner iterations.

    References:
        Ying, L., & Sheng, J. (2007).
        Joint image reconstruction and sensitivity estimation in SENSE
        (JSENSE).
        Magnetic Resonance in Medicine, 57(6), 1196-1202.

        Uecker, M., Hohage, T., Block, K. T., & Frahm, J. (2008).
        Image reconstruction by regularized nonlinear inversion-
        joint estimation of coil sensitivities and image content.
        Magnetic Resonance in Medicine, 60(#), 674-682.

    """

    def __init__(
        self,
        y,
        mps_ker_width=16,
        ksp_calib_width=24,
        lamda=0,
        device=sp.cpu_device,
        comm=None,
        weights=None,
        coord=None,
        img_shape=None,
        grd_shape=None,
        max_iter=10,
        max_inner_iter=10,
        normalize=True,
        show_pbar=True,
    ):
        self.y = y
        self.mps_ker_width = mps_ker_width
        self.ksp_calib_width = ksp_calib_width
        self.lamda = lamda
        self.weights = weights
        self.coord = coord
        self.img_shape = img_shape
        self.grd_shape = grd_shape
        self.max_iter = max_iter
        self.max_inner_iter = max_inner_iter
        self.normalize = normalize

        self.device = sp.Device(device)
        self.comm = comm
        self.dtype = y.dtype
        self.num_coils = len(y)
        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        self._get_data()
        self._get_vars()
        self._get_alg()
        super().__init__(self.alg, show_pbar=show_pbar)

    def _get_data(self):
        if self.coord is None:
            self.img_shape = list(self.y.shape[1:])
            ndim = len(self.img_shape)

            self.y = sp.resize(
                self.y, [self.num_coils] + ndim * [self.ksp_calib_width]
            )

            if self.weights is not None:
                self.weights = sp.resize(
                    self.weights, ndim * [self.ksp_calib_width]
                )

        else:
            if self.img_shape is None:
                self.img_shape = sp.estimate_shape(self.coord)
            else:
                self.img_shape = list(self.img_shape)

            calib_idx = (
                np.amax(np.abs(self.coord), axis=-1) < self.ksp_calib_width / 2
            )

            self.coord = self.coord[calib_idx]
            self.y = self.y[:, calib_idx]

            if self.weights is not None:
                self.weights = self.weights[calib_idx]

        if self.weights is None:
            self.y = sp.to_device(self.y, self.device)
        else:
            self.y = sp.to_device(self.weights**0.5 * self.y, self.device)

        if self.coord is not None:
            self.coord = sp.to_device(self.coord, self.device)
        if self.weights is not None:
            self.weights = sp.to_device(self.weights, self.device)

        self.weights = _estimate_weights(self.y, self.weights, self.coord)

        if self.normalize:
            xp = self.device.xp
            with self.device:
                self.y = self.y / xp.linalg.norm(self.y)

    def _get_vars(self):
        ndim = len(self.img_shape)

        mps_ker_shape = [self.num_coils] + [self.mps_ker_width] * ndim
        if self.coord is None:
            img_ker_shape = [
                i + self.mps_ker_width - 1 for i in self.y.shape[1:]
            ]
        else:
            if self.grd_shape is None:
                self.grd_shape = sp.estimate_shape(self.coord)

            img_ker_shape = [
                i + self.mps_ker_width - 1 for i in self.grd_shape
            ]

        self.img_ker = sp.dirac(
            img_ker_shape, dtype=self.dtype, device=self.device
        )
        with self.device:
            self.mps_ker = self.device.xp.zeros(
                mps_ker_shape, dtype=self.dtype
            )

    def _get_alg(self):
        def min_mps_ker():
            self.A_mps_ker = linop.ConvImage(
                self.mps_ker.shape,
                self.img_ker,
                coord=self.coord,
                weights=self.weights,
            )
            sp.app.LinearLeastSquares(
                self.A_mps_ker,
                self.y,
                self.mps_ker,
                lamda=self.lamda,
                max_iter=self.max_inner_iter,
                show_pbar=False,
            ).run()

        def min_img_ker():
            self.A_img_ker = linop.ConvSense(
                self.img_ker.shape,
                self.mps_ker,
                coord=self.coord,
                weights=self.weights,
                comm=self.comm,
            )
            sp.app.LinearLeastSquares(
                self.A_img_ker,
                self.y,
                self.img_ker,
                lamda=self.lamda,
                max_iter=self.max_inner_iter,
                show_pbar=False,
            ).run()

        self.alg = sp.alg.AltMin(
            min_mps_ker, min_img_ker, max_iter=self.max_iter
        )

    def _output(self):
        xp = self.device.xp
        # Normalize by root-sum-of-squares.
        with self.device:
            rss = 0
            mps = np.empty([self.num_coils] + self.img_shape, dtype=self.dtype)
            for c in range(self.num_coils):
                mps_c = sp.ifft(sp.resize(self.mps_ker[c], self.img_shape))
                rss += xp.abs(mps_c) ** 2
                sp.copyto(mps[c], mps_c)

            rss = sp.to_device(rss)
            if self.comm is not None:
                self.comm.allreduce(rss)

            rss = rss**0.5
            mps /= rss
            return mps


class EspiritCalib(sp.app.App):
    """ESPIRiT calibration.

    Currently only supports outputting one set of maps.

    Args:
        ksp (array): k-space array of shape [num_coils, n_ndim, ..., n_1]
        calib (tuple of ints): length-2 image shape.
        thresh (float): threshold for the calibration matrix.
        kernel_width (int): kernel width for the calibration matrix.
        max_power_iter (int): maximum number of power iterations.
        device (Device): computing device.
        crop (int): cropping threshold.

    Returns:
        array: ESPIRiT maps of the same shape as ksp.

    References:
        Martin Uecker, Peng Lai, Mark J. Murphy, Patrick Virtue, Michael Elad,
        John M. Pauly, Shreyas S. Vasanawala, and Michael Lustig
        ESPIRIT - An Eigenvalue Approach to Autocalibrating Parallel MRI:
        Where SENSE meets GRAPPA.
        Magnetic Resonance in Medicine, 71:990-1001 (2014)

    """

    def __init__(
        self,
        ksp,
        calib_width=24,
        thresh=0.02,
        kernel_width=6,
        crop=0.95,
        max_iter=100,
        device=sp.cpu_device,
        output_eigenvalue=False,
        show_pbar=True,
    ):
        self.device = sp.Device(device)
        self.output_eigenvalue = output_eigenvalue
        self.crop = crop

        img_ndim = ksp.ndim - 1
        num_coils = len(ksp)
        with sp.get_device(ksp):
            # Get calibration region
            calib_shape = [num_coils] + [calib_width] * img_ndim
            calib = sp.resize(ksp, calib_shape)
            calib = sp.to_device(calib, device)

        xp = self.device.xp
        with self.device:
            # Get calibration matrix.
            # Shape [num_coils] + num_blks + [kernel_width] * img_ndim
            mat = sp.array_to_blocks(
                calib, [kernel_width] * img_ndim, [1] * img_ndim
            )
            mat = mat.reshape([num_coils, -1, kernel_width**img_ndim])
            mat = mat.transpose([1, 0, 2])
            mat = mat.reshape([-1, num_coils * kernel_width**img_ndim])

            # Perform SVD on calibration matrix
            _, S, VH = xp.linalg.svd(mat, full_matrices=False)
            VH = VH[S > thresh * S.max(), :]

            # Get kernels
            num_kernels = len(VH)
            kernels = VH.reshape(
                [num_kernels, num_coils] + [kernel_width] * img_ndim
            )
            img_shape = ksp.shape[1:]

            # Get covariance matrix in image domain
            AHA = xp.zeros(
                img_shape[::-1] + (num_coils, num_coils), dtype=ksp.dtype
            )
            for kernel in kernels:
                img_kernel = sp.ifft(
                    sp.resize(kernel, ksp.shape), axes=range(-img_ndim, 0)
                )
                aH = xp.expand_dims(img_kernel.T, axis=-1)
                a = xp.conj(aH.swapaxes(-1, -2))
                AHA += aH @ a

            AHA *= sp.prod(img_shape) / kernel_width**img_ndim
            self.mps = xp.ones(ksp.shape[::-1] + (1,), dtype=ksp.dtype)

            def forward(x):
                with sp.get_device(x):
                    return AHA @ x

            def normalize(x):
                with sp.get_device(x):
                    return (
                        xp.sum(xp.abs(x) ** 2, axis=-2, keepdims=True) ** 0.5
                    )

            alg = sp.alg.PowerMethod(
                forward, self.mps, norm_func=normalize, max_iter=max_iter
            )

        super().__init__(alg, show_pbar=show_pbar)

    def _output(self):
        xp = self.device.xp
        with self.device:
            # Normalize phase with respect to first channel
            mps = self.mps.T[0]
            mps *= xp.conj(mps[0] / xp.abs(mps[0]))

            # Crop maps by thresholding eigenvalue
            max_eig = self.alg.max_eig.T[0]
            mps *= max_eig > self.crop

        if self.output_eigenvalue:
            return mps, max_eig
        else:
            return mps



# -*- coding: utf-8 -*-
"""MRI linear operators.

This module mainly contains the Sense linear operator,
which integrates multi-channel coil sensitivity maps and
discrete Fourier transform.

"""
import sigpy as sp


def Sense(
    mps,
    coord=None,
    weights=None,
    tseg=None,
    ishape=None,
    coil_batch_size=None,
    comm=None,
    transp_nufft=False,
):
    """Sense linear operator.

    Args:
        mps (array): sensitivity maps of length = number of channels.
        coord (None or array): coordinates.
        weights (None or array): k-space weights.
            Useful for soft-gating or density compensation.
        tseg (None or Dictionary): parameters for time-segmented off-resonance
            correction. Parameters are 'b0' (array), 'dt' (float),
            'lseg' (int), and 'n_bins' (int). Lseg is the number of
            time segments used, and n_bins is the number of histogram bins.
        ishape (None or tuple): image shape.
        coil_batch_size (None or int): batch size for processing multi-channel.
            When None, process all coils at the same time.
            Useful for saving memory.
        comm (None or `sigpy.Communicator`): communicator
            for distributed computing.

    """
    # Get image shape and dimension.
    num_coils = len(mps)
    if ishape is None:
        ishape = mps.shape[1:]
        img_ndim = mps.ndim - 1
    else:
        img_ndim = len(ishape)

    # Serialize linop if coil_batch_size is smaller than num_coils.
    num_coils = len(mps)
    if coil_batch_size is None:
        coil_batch_size = num_coils

    if coil_batch_size < len(mps):
        num_coil_batches = (num_coils + coil_batch_size - 1) // coil_batch_size
        A = sp.linop.Vstack(
            [
                Sense(
                    mps[c * coil_batch_size : ((c + 1) * coil_batch_size)],
                    coord=coord,
                    weights=weights,
                    ishape=ishape,
                )
                for c in range(num_coil_batches)
            ],
            axis=0,
        )

        if comm is not None:
            C = sp.linop.AllReduceAdjoint(ishape, comm, in_place=True)
            A = A * C

        return A

    # Create Sense linear operator
    S = sp.linop.Multiply(ishape, mps)
    if tseg is None:
        if coord is None:
            F = sp.linop.FFT(S.oshape, axes=range(-img_ndim, 0))
        else:
            if transp_nufft is False:
                F = sp.linop.NUFFT(S.oshape, coord)
            else:
                F = sp.linop.NUFFT(S.oshape, -coord).H

        A = F * S

    # If B0 provided, perform time-segmented off-resonance compensation
    else:
        if transp_nufft is False:
            F = sp.linop.NUFFT(S.oshape, coord)
        else:
            F = sp.linop.NUFFT(S.oshape, -coord).H
        time = len(coord) * tseg["dt"]
        b, ct = sp.mri.util.tseg_off_res_b_ct(
            tseg["b0"], tseg["n_bins"], tseg["lseg"], tseg["dt"], time
        )
        for ii in range(tseg["lseg"]):
            Bi = sp.linop.Multiply(F.oshape, b[:, ii])
            Cti = sp.linop.Multiply(S.ishape, ct[:, ii].reshape(S.ishape))

            # operation below is effectively A = A + Bi * F(Cti * S)
            if ii == 0:
                A = Bi * F * S * Cti
            else:
                A = A + Bi * F * S * Cti

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(F.oshape, weights**0.5)

        A = P * A

    if comm is not None:
        C = sp.linop.AllReduceAdjoint(ishape, comm, in_place=True)
        A = A * C

    A.repr_str = "Sense"
    return A


def ConvSense(
    img_ker_shape, mps_ker, coord=None, weights=None, grd_shape=None, comm=None
):
    """Convolution linear operator with sensitivity maps kernel in k-space.

    Args:
        img_ker_shape (tuple of ints): image kernel shape.
        mps_ker (array): sensitivity maps kernel.
        coord (array): coordinates.
        grd_shape (None or list): Shape of grid.

    """
    ndim = len(img_ker_shape)
    num_coils = mps_ker.shape[0]
    mps_ker = mps_ker.reshape((num_coils, 1) + mps_ker.shape[1:])
    R = sp.linop.Reshape((1,) + tuple(img_ker_shape), img_ker_shape)
    C = sp.linop.ConvolveData(
        R.oshape, mps_ker, mode="valid", multi_channel=True
    )
    A = C * R

    if coord is not None:
        if grd_shape is None:
            grd_shape = sp.estimate_shape(coord)
        else:
            grd_shape = list(grd_shape)

        grd_shape = [num_coils] + grd_shape
        iF = sp.linop.IFFT(grd_shape, axes=range(-ndim, 0))
        N = sp.linop.NUFFT(grd_shape, coord)
        A = N * iF * A

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(A.oshape, weights**0.5)

        A = P * A

    if comm is not None:
        C = sp.linop.AllReduceAdjoint(img_ker_shape, comm, in_place=True)
        A = A * C

    return A


def ConvImage(
    mps_ker_shape, img_ker, coord=None, weights=None, grd_shape=None
):
    """Convolution linear operator with image kernel in k-space.

    Args:
        mps_ker_shape (tuple of ints): sensitivity maps kernel shape.
        img_ker (array): image kernel.
        coord (array): coordinates.
        grd_shape (None or list): Shape of grid.

    """
    ndim = img_ker.ndim
    num_coils = mps_ker_shape[0]
    img_ker = img_ker.reshape((1,) + img_ker.shape)
    R = sp.linop.Reshape(
        (num_coils, 1) + tuple(mps_ker_shape[1:]), mps_ker_shape
    )
    C = sp.linop.ConvolveFilter(
        R.oshape, img_ker, mode="valid", multi_channel=True
    )
    A = C * R

    if coord is not None:
        num_coils = mps_ker_shape[0]
        if grd_shape is None:
            grd_shape = sp.estimate_shape(coord)
        else:
            grd_shape = list(grd_shape)

        grd_shape = [num_coils] + grd_shape
        iF = sp.linop.IFFT(grd_shape, axes=range(-ndim, 0))
        N = sp.linop.NUFFT(grd_shape, coord)
        A = N * iF * A

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(A.oshape, weights**0.5)
        A = P * A

    return A



# -*- coding: utf-8 -*-
"""Adiabatic Pulse Design functions.

"""
import numpy as np

__all__ = ["bir4", "hypsec", "wurst", "goia_wurst", "bloch_siegert_fm"]


def bir4(n, beta, kappa, theta, dw0):
    r"""Design a BIR-4 adiabatic pulse.

    BIR-4 is equivalent to two BIR-1 pulses back-to-back.

    Args:
        n (int): number of samples (should be a multiple of 4).
        beta (float): AM waveform parameter.
        kappa (float): FM waveform parameter.
        theta (float): flip angle in radians.
        dw0: FM waveform scaling (radians/s).

    Returns:
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References:
        Staewen, R.S. et al. (1990). '3-D FLASH Imaging using a single surface
        coil and a new adiabatic pulse, BIR-4'.
        Invest. Radiology, 25:559-567.
    """

    dphi = np.pi + theta / 2

    t = np.arange(0, n) / n

    a1 = np.tanh(beta * (1 - 4 * t[: n // 4]))
    a2 = np.tanh(beta * (4 * t[n // 4 : n // 2] - 1))
    a3 = np.tanh(beta * (3 - 4 * t[n // 2 : 3 * n // 4]))
    a4 = np.tanh(beta * (4 * t[3 * n // 4 :] - 3))

    a = np.concatenate((a1, a2, a3, a4)).astype(np.complex64)
    a[n // 4 : 3 * n // 4] = a[n // 4 : 3 * n // 4] * np.exp(1j * dphi)

    om1 = dw0 * np.tan(kappa * 4 * t[: n // 4]) / np.tan(kappa)
    om2 = dw0 * np.tan(kappa * (4 * t[n // 4 : n // 2] - 2)) / np.tan(kappa)
    om3 = (
        dw0 * np.tan(kappa * (4 * t[n // 2 : 3 * n // 4] - 2)) / np.tan(kappa)
    )
    om4 = dw0 * np.tan(kappa * (4 * t[3 * n // 4 :] - 4)) / np.tan(kappa)

    om = np.concatenate((om1, om2, om3, om4))

    return a, om


def hypsec(n=512, beta=800, mu=4.9, dur=0.012):
    r"""Design a hyperbolic secant adiabatic pulse.

    mu * beta becomes the amplitude of the frequency sweep

    Args:
        n (int): number of samples (should be a multiple of 4).
        beta (float): AM waveform parameter.
        mu (float): a constant, determines amplitude of frequency sweep.
        dur (float): pulse time (s).

    Returns:
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References:
        Baum, J., Tycko, R. and Pines, A. (1985). 'Broadband and adiabatic
        inversion of a two-level system by phase-modulated pulses'.
        Phys. Rev. A., 32:3435-3447.
    """

    t = np.arange(-n // 2, n // 2) / n * dur

    a = np.cosh(beta * t) ** -1
    om = -mu * beta * np.tanh(beta * t)

    return a, om


def wurst(n=512, n_fac=40, bw=40e3, dur=2e-3):
    r"""Design a WURST (wideband, uniform rate, smooth truncation) adiabatic
     inversion pulse

    Args:
        n (int): number of samples (should be a multiple of 4).
        n_fac (int): power to exponentiate to within AM term. ~20 or greater is
         typical.
        bw (float): pulse bandwidth.
        dur (float): pulse time (s).


    Returns:
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References:
        Kupce, E. and Freeman, R. (1995). 'Stretched Adiabatic Pulses for
        Broadband Spin Inversion'.
        J. Magn. Reson. Ser. A., 117:246-256.
    """

    t = np.arange(0, n) * dur / n

    a = 1 - np.power(np.abs(np.cos(np.pi * t / dur)), n_fac)
    om = np.linspace(-bw / 2, bw / 2, n) * 2 * np.pi

    return a, om


def goia_wurst(
    n=512, dur=3.5e-3, f=0.9, n_b1=16, m_grad=4, b1_max=817, bw=20000
):
    r"""Design a GOIA (gradient offset independent adiabaticity) WURST
     inversion pulse

    Args:
        n (int): number of samples.
        dur (float): pulse duration (s).
        f (float): [0,1] gradient modulation factor
        n_b1 (int): order for B1 modulation
        m_grad (int): order for gradient modulation
        b1_max (float): maximum b1 (Hz)
        bw (float): pulse bandwidth (Hz)

    Returns:
        3-element tuple containing:

        - **a** (*array*): AM waveform (Hz)
        - **om** (*array*): FM waveform (Hz)
        - **g** (*array*): normalized gradient waveform

    References:
        O. C. Andronesi, S. Ramadan, E.-M. Ratai, D. Jennings, C. E. Mountford,
        A. G. Sorenson.
        J Magn Reson, 203:283-293, 2010.

    """

    t = np.arange(0, n) * dur / n

    a = b1_max * (1 - np.abs(np.sin(np.pi / 2 * (2 * t / dur - 1))) ** n_b1)
    g = (1 - f) + f * np.abs(np.sin(np.pi / 2 * (2 * t / dur - 1))) ** m_grad
    om = np.cumsum((a**2) / g) * dur / n
    om = om - om[n // 2 + 1]
    om = g * om
    om = om / np.max(np.abs(om)) * bw / 2

    return a, om, g


def bloch_siegert_fm(
    n=512, dur=2e-3, b1p=20.0, k=42.0, gamma=2 * np.pi * 42.58
):
    r"""
    U-shaped FM waveform for adiabatic Bloch-Siegert :math:`B_1^{+}` mapping
    and spatial encoding.

    Args:
        n (int): number of time points
        dur (float): duration in seconds
        b1p (float): nominal amplitude of constant AM waveform
        k (float): design parameter that affects max in-band
            perturbation
        gamma (float): gyromagnetic ratio

    Returns:
        om (array): FM waveform (radians/s).

    References:
        M. M. Khalighi, B. K. Rutt, and A. B. Kerr.
        Adiabatic RF pulse design for Bloch-Siegert B1+ mapping.
        Magn Reson Med, 70(3):829–835, 2013.

        M. Jankiewicz, J. C. Gore, and W. A. Grissom.
        Improved encoding pulses for Bloch-Siegert B1+ mapping.
        J Magn Reson, 226:79–87, 2013.

    """

    t = np.arange(1, n // 2) * dur / n

    om = gamma * b1p / np.sqrt((1 - gamma * b1p / k * t) ** -2 - 1)
    om = np.concatenate((om, om[::-1]))

    return om



import pickle
import unittest

import numpy as np
import numpy.testing as npt

from sigpy import backend, config, linop, util

if __name__ == "__main__":
    unittest.main()

dtypes = [np.float32, np.float64, np.complex64, np.complex128]
devices = [backend.cpu_device]
if config.cupy_enabled:
    devices.append(backend.Device(0))


class TestLinop(unittest.TestCase):
    def check_linop_unitary(self, A, device=backend.cpu_device, dtype=float):
        device = backend.Device(device)
        x = util.randn(A.ishape, dtype=dtype, device=device)
        xp = device.xp
        with device:
            xp.testing.assert_allclose(A.H * A * x, x, atol=1e-5, rtol=1e-5)

    def check_linop_linear(self, A, device=backend.cpu_device, dtype=float):
        device = backend.Device(device)
        a = util.randn([1], dtype=dtype, device=device)
        x = util.randn(A.ishape, dtype=dtype, device=device)
        y = util.randn(A.ishape, dtype=dtype, device=device)

        xp = device.xp
        with device:
            xp.testing.assert_allclose(
                A(a * x + y), a * A(x) + A(y), atol=1e-5, rtol=1e-5
            )

    def check_linop_adjoint(self, A, device=backend.cpu_device, dtype=float):
        device = backend.Device(device)
        x = util.randn(A.ishape, dtype=dtype, device=device)
        y = util.randn(A.oshape, dtype=dtype, device=device)

        xp = device.xp
        with device:
            lhs = xp.vdot(A * x, y)
            rhs = xp.vdot(A.H.H * x, y)
            xp.testing.assert_allclose(lhs, rhs, atol=1e-5, rtol=1e-5)

            rhs = xp.vdot(x, A.H * y)
            xp.testing.assert_allclose(lhs, rhs, rtol=1e-3)

    def check_linop_normal(self, A, device=backend.cpu_device, dtype=float):
        device = backend.Device(device)
        x = util.randn(A.ishape, dtype=dtype, device=device)

        xp = device.xp
        with device:
            lhs = A.H * A * x
            rhs = A.N * x
            xp.testing.assert_allclose(lhs, rhs, atol=1e-2, rtol=1e-3)

    def check_linop_pickleable(self, A):
        with self.subTest(A=A):
            assert A.__repr__() == pickle.loads(pickle.dumps(A)).__repr__()

    def test_Identity(self):
        shape = [5]
        A = linop.Identity(shape)
        x = util.randn(shape)

        npt.assert_allclose(A(x), x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_unitary(A)
        self.check_linop_pickleable(A)

    def test_Conj(self):
        shape = [5]
        Id = linop.Identity(shape)
        A = linop.Conj(Id)
        x = util.randn(shape)

        npt.assert_allclose(A(x), x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Add(self):
        shape = [5]
        Id = linop.Identity(shape)
        A = linop.Add([Id, Id])
        x = util.randn(shape)

        npt.assert_allclose(A(x), 2 * x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Compose(self):
        shape = [5]
        Id = linop.Identity(shape)
        A = linop.Compose([Id, Id])
        x = util.randn(shape)

        npt.assert_allclose(A(x), x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Hstack(self):
        shape = [5]
        Id = linop.Identity(shape)
        x1 = util.randn(shape)
        x2 = util.randn(shape)
        x = util.vec([x1, x2])

        A = linop.Hstack([Id, Id])
        npt.assert_allclose(A(x), x1 + x2)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        shape = [5, 3]
        Id = linop.Identity(shape)
        x1 = util.randn(shape)
        x2 = util.randn(shape)
        x = np.concatenate([x1, x2], axis=1)

        A = linop.Hstack([Id, Id], axis=1)
        npt.assert_allclose(A(x), x1 + x2)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Vstack(self):
        shape = [5]
        Id = linop.Identity(shape)
        x = util.randn(shape)

        A = linop.Vstack([Id, Id])
        npt.assert_allclose(A(x), util.vec([x, x]))
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        shape = [5, 3]
        Id = linop.Identity(shape)
        x = util.randn(shape)

        A = linop.Vstack([Id, Id], axis=1)
        npt.assert_allclose(A(x), np.concatenate([x, x], axis=1))
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Diag(self):
        shape = [5]
        Id = linop.Identity(shape)
        x = util.randn([10])

        A = linop.Diag([Id, Id])
        npt.assert_allclose(A(x), x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        shape = [5, 3]
        Id = linop.Identity(shape)
        x = util.randn([5, 6])

        A = linop.Diag([Id, Id], iaxis=1, oaxis=1)
        npt.assert_allclose(A(x), x)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_FFT(self):
        for ndim in [1, 2, 3]:
            for n in [3, 4, 5, 6]:
                ishape = [n] * ndim
                A = linop.FFT(ishape)
                self.check_linop_linear(A)
                self.check_linop_adjoint(A)
                self.check_linop_normal(A)
                self.check_linop_unitary(A)
                self.check_linop_pickleable(A)

    def test_NUFFT(self):
        for ndim in [1, 2, 3]:
            for n in [2, 3, 4, 5, 6]:
                ishape = [3] * ndim
                coord = np.random.random([10, ndim])

                A = linop.NUFFT(ishape, coord, toeplitz=True, oversamp=2)
                self.check_linop_linear(A)
                self.check_linop_adjoint(A)
                self.check_linop_normal(A)
                self.check_linop_pickleable(A)

    def test_MatMul(self):
        mshape = (5, 4, 2)
        ishape = (5, 2, 3)
        A = linop.MatMul(ishape, util.randn(mshape))
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_RightMatMul(self):
        ishape = (5, 4, 2)
        mshape = (5, 2, 3)
        A = linop.RightMatMul(ishape, util.randn(mshape))
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_Multiply(self):
        # Test scalar
        ishape = [2]
        mult = 1.1

        A = linop.Multiply(ishape, mult)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        x = np.array([1.0, 2.0], complex)
        y = np.array([1.1, 2.2], complex)
        npt.assert_allclose(A * x, y)

        # Test simple
        ishape = [2]
        mult = np.array([1.0, 2.0])

        A = linop.Multiply(ishape, mult)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        x = np.array([1.0, 2.0], complex)
        y = np.array([1.0, 4.0], complex)
        npt.assert_allclose(A * x, y)

        # Test broadcasting
        ishape = [2]
        mult = np.array([[1.0, 2.0], [3.0, 4.0]])

        A = linop.Multiply(ishape, mult)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        x = np.array([1.0, 2.0], complex)
        y = np.array([[1.0, 4.0], [3.0, 8.0]], complex)
        npt.assert_allclose(A * x, y)

    def test_Resize(self):
        ishape = [3]
        oshape = [5]

        A = linop.Resize(oshape, ishape)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        A = linop.Resize(oshape, ishape, oshift=[1])
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        ishape = [5]
        oshape = [3]

        A = linop.Resize(oshape, ishape)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        A = linop.Resize(oshape, ishape, ishift=[1])
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Downsample(self):
        ishape = [5]
        factors = [2]

        A = linop.Downsample(ishape, factors)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        A = linop.Downsample(ishape, factors, shift=[1])
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Upsample(self):
        oshape = [5]
        factors = [2]

        A = linop.Downsample(oshape, factors)
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

        A = linop.Downsample(oshape, factors, shift=[1])
        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)

    def test_Interpolate(self):
        # Test linear interpolation.
        ishape = [5]
        coord = np.array([[0.5], [1.5], [2.5]])

        A = linop.Interpolate(ishape, coord)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        x = np.array([0.0, 1.0, 2.0, 1.0, 0.0])

        npt.assert_allclose([0.5, 1.5, 1.5], A * x)

        # Test no batch
        ishape = [2, 2]
        coord = np.array([[0, 0], [1, 0], [1.5, 0]]) / 4.0

        A = linop.Interpolate(ishape, coord)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        # Test batch
        ishape = [2, 2, 2]

        A = linop.Interpolate(ishape, coord)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_Wavelet(self):
        shape = [16]
        A = linop.Wavelet(shape, level=1, wave_name="haar")
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_unitary(A)
        self.check_linop_pickleable(A)

        shape = [129]
        A = linop.Wavelet(shape)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_unitary(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        shape = [17]
        A = linop.Wavelet(shape, level=1)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_unitary(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_Circshift(self):
        shape = [8]
        shift = [4]
        A = linop.Circshift(shape, shift)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_unitary(A)
        self.check_linop_pickleable(A)

    def test_FiniteDifference(self):
        shape = [8]
        A = linop.FiniteDifference(shape)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_Transpose(self):
        shape = [3, 4]
        A = linop.Transpose(shape)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_unitary(A)
        self.check_linop_pickleable(A)

        x = util.randn(shape)
        npt.assert_allclose(A(x), np.transpose(x))

    def test_Sum(self):
        shape = [2, 3, 4, 2]
        axes = [1, 3]
        A = linop.Sum(shape, axes)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

    def test_ArrayToBlocks(self):
        ishape = [4]
        blk_shape = [2]
        blk_strides = [2]

        A = linop.ArrayToBlocks(ishape, blk_shape, blk_strides)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_linear(A)
        self.check_linop_pickleable(A)

        x = np.array([1, 2, 3, 4], complex)

        npt.assert_allclose(A(x), [[1, 2], [3, 4]])

    def test_ConvolveData(self):
        for device in devices:
            for dtype in dtypes:
                for mode in ["full", "valid"]:
                    with self.subTest(mode=mode, dtype=dtype, device=device):
                        data_shape = [3, 4]
                        filt = util.randn([2, 3], dtype=dtype)
                        A = linop.ConvolveData(data_shape, filt, mode=mode)
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        data_shape = [4, 3, 4]
                        filt = util.randn([1, 4, 2, 3], dtype=dtype)
                        A = linop.ConvolveData(
                            data_shape, filt, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        data_shape = [1, 3, 4]
                        filt = util.randn([4, 1, 2, 3], dtype=dtype)
                        A = linop.ConvolveData(
                            data_shape, filt, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        data_shape = [2, 3, 4]
                        filt = util.randn([4, 2, 2, 3], dtype=dtype)
                        A = linop.ConvolveData(
                            data_shape, filt, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        data_shape = [2, 3, 4]
                        filt = util.randn([4, 2, 2, 3], dtype=dtype)
                        strides = [2, 2]
                        A = linop.ConvolveData(
                            data_shape,
                            filt,
                            mode=mode,
                            strides=strides,
                            multi_channel=True,
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

    def test_ConvolveFilter(self):
        for device in devices:
            for dtype in dtypes:
                for mode in ["full", "valid"]:
                    with self.subTest(mode=mode, dtype=dtype, device=device):
                        filt_shape = [2, 3]
                        data = util.randn([3, 4], dtype=dtype)
                        A = linop.ConvolveFilter(filt_shape, data, mode=mode)
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        filt_shape = [1, 4, 2, 3]
                        data = util.randn([4, 3, 4], dtype=dtype)
                        A = linop.ConvolveFilter(
                            filt_shape, data, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        filt_shape = [4, 1, 2, 3]
                        data = util.randn([1, 3, 4], dtype=dtype)
                        A = linop.ConvolveFilter(
                            filt_shape, data, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        filt_shape = [4, 2, 2, 3]
                        data = util.randn([2, 3, 4], dtype=dtype)
                        A = linop.ConvolveFilter(
                            filt_shape, data, mode=mode, multi_channel=True
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

                        filt_shape = [4, 2, 2, 3]
                        strides = [2, 2]
                        data = util.randn([2, 3, 4], dtype=dtype)
                        A = linop.ConvolveFilter(
                            filt_shape,
                            data,
                            mode=mode,
                            strides=strides,
                            multi_channel=True,
                        )
                        self.check_linop_linear(A, dtype=dtype, device=device)
                        self.check_linop_adjoint(A, dtype=dtype, device=device)
                        self.check_linop_normal(A, dtype=dtype, device=device)
                        self.check_linop_pickleable(A)

    def test_Slice(self):
        ishape = (5,)
        idx = slice(1, 3)
        A = linop.Slice(ishape, idx)
        x = np.arange(5)
        npt.assert_allclose(A(x), [1, 2])

        ishape = (5, 5)
        idx = (slice(1, 3), slice(1, 3))
        A = linop.Slice(ishape, idx)
        x = np.outer(np.arange(5), np.arange(5))
        npt.assert_allclose(A(x), [[1, 2], [2, 4]])

        self.check_linop_linear(A)
        self.check_linop_adjoint(A)
        self.check_linop_normal(A)
        self.check_linop_pickleable(A)



# -*- coding: utf-8 -*-
"""MRI waveform import/export files.
"""

import struct

import numpy as np

__all__ = ["signa", "ge_rf_params", "philips_rf_params", "siemens_rf"]


def siemens_rf(
    pulse, rfbw, rfdurms, pulsename, minslice=0.5, maxslice=320.0, comment=None
):
    """Write a .pta text file for Siemens PulseTool.

    Args:
        pulse (array): complex-valued RF pulse array with maximum of 4096
            points.
        rfbw (float): bandwidth of RF pulse in Hz
        rfdurms (float): duration of RF pulse in ms
        pulsename (string): '<FamilyName>.<PulseName>', e.g. 'Sigpy.SincPulse'
        minslice (float): minimum slice thickness [mm]
        maxslice (float): maximum slice thickness [mm]
        comment (string): a comment that can be seen in Siemens PulseTool

    Note this has only been tested on MAGNETOM Verio running (VB17)

    Open pulsetool from the IDEA command line. Open the extrf.dat file and add
    this .pta file using the import function

    Recommended to make a copy and renaming extrf.dat prior to making changes.

    After saving a new pulse to <myUniqueFileName>_extrf.dat and copying it to
    the scanner, you will need to re-boot the host for it to load changes.

    """

    # get the number of points in RF waveform
    npts = pulse.size
    assert npts <= 4096, (
        "RF pulse must have less than 4096 points for" " Siemens VB17"
    )

    if comment is None:
        comment = ""

    # Calculate reference gradient value.
    # This is necessary for proper calculation of slice-select gradient
    # amplitude using the .getGSAmplitude() method for the external RF class.
    # See the IDEA documentation for more details on this.
    refgrad = 1000.0 * rfbw * (rfdurms / 5.12) / (42.577e06 * (10.0 / 1000.0))

    rffile = open(pulsename + ".pta", "w")
    rffile.write("PULSENAME: {}\n".format(pulsename))
    rffile.write("COMMENT: {}\n".format(comment))
    rffile.write("REFGRAD: {:6.5f}\n".format(refgrad))
    rffile.write("MINSLICE: {:6.5f}\n".format(minslice))
    rffile.write("MAXSLICE: {:6.5f}\n".format(maxslice))

    # the following are related to SAR calcs and will be calculated by
    # PulseTool upon loading the pulse
    rffile.write("AMPINT: \n")
    rffile.write("POWERINT: \n")
    rffile.write("ABSINT: \n\n")

    # magnitude must be between 0 and 1
    mxmag = np.max(np.abs(pulse))
    for n in range(npts):
        mag = np.abs(pulse[n]) / mxmag  # magnitude at current point
        mag = np.squeeze(mag)
        pha = np.angle(pulse[n])  # phase at current point
        pha = np.squeeze(pha)
        rffile.write("{:10.9f}\t{:10.9f}\t; ({:d})\n".format(mag, pha, n))
    rffile.close()


def signa(wav, filename, scale=-1):
    """Write a binary waveform in the GE format.

    Args:
        wav (array): waveform (gradient or RF), may be complex-valued.
        filename (string): filename to write to.
        scale (float): scaling factor to apply (default = waveform's max)

    Adapted from John Pauly's RF Tools signa() MATLAB function

    """

    wmax = int("7ffe", 16)

    if not np.iscomplexobj(wav):
        if scale == -1:
            scale = 1 / np.max(np.abs(wav))

        # scale up to fit in a short integer
        wav = wav * scale * wmax

        # mask off low bit, since it would be an EOS otherwise
        wav = 2 * np.round(wav / 2)

        fid = open(filename, "wb")

        for x in np.nditer(wav):
            fid.write(struct.pack(">h", int(x.item())))

        fid.close()

    else:
        if scale == -1:
            scale = 1 / np.max(
                (np.max(np.abs(np.real(wav))), np.max(np.abs(np.imag(wav))))
            )

        # scale up to fit in a short integer
        wav = wav * scale * wmax

        # mask off low bit, since it would be an EOS otherwise
        wav = 2 * np.round(wav / 2)

        fid = open(filename + ".r", "wb")

        for x in np.nditer(wav):
            fid.write(struct.pack(">h", int(np.real(x))))

        fid.close()

        fid = open(filename + ".i", "wb")

        for x in np.nditer(wav):
            fid.write(struct.pack(">h", int(np.imag(x))))

        fid.close()


def ge_rf_params(rf, dt=4e-6):
    """Calculate RF pulse parameters for deployment
    on a GE scanner.

    Args:
        rf (array): RF pulse samples
        dt (scalar): RF dwell time (seconds)

    Adapted from Adam Kerr's rf_save() MATLAB function

    """

    print("GE RF Pulse Parameters:")

    n = len(rf)
    rfn = rf / np.max(np.abs(rf))

    abswidth = np.sum(np.abs(rfn)) / n
    print("abswidth = ", abswidth)

    effwidth = np.sum(np.abs(rfn) ** 2) / n
    print("effwidth = ", effwidth)

    print("area = ", abswidth)

    pon = np.abs(rfn) > 0
    temp_pw = 0
    max_pw = 0
    for i in range(0, len(rfn)):
        if pon[i] == 0 & temp_pw > 0:
            max_pw = np.max(max_pw, temp_pw)
            temp_pw = 0
    max_pw = max_pw / n

    dty_cyc = np.sum(np.abs(rfn) > 0.2236) / n
    if dty_cyc < max_pw:
        dty_cyc = max_pw
    print("dtycyc = ", dty_cyc)
    print("maxpw = ", max_pw)

    max_b1 = np.max(np.abs(rf))
    print("max_b1 = ", max_b1)

    int_b1_sqr = np.sum(np.abs(rf) ** 2) * dt * 1e3
    print("int_b1_sqr = ", int_b1_sqr)

    rms_b1 = np.sqrt(np.sum(np.abs(rf) ** 2)) / n
    print("max_rms_b1 = ", rms_b1)


def philips_rf_params(rf):
    """Calculate RF pulse parameters for deployment
    on a Philips scanner.

    Args:
        rf (array): RF pulse samples (assumed real-valued)

    """

    print("Philips RF Pulse Parameters")

    n = len(rf)
    rfn = rf / np.max(np.abs(rf))

    am_c_teff = np.sum(rfn * 32767) / (32767 * n)
    print("am_c_teff = ", am_c_teff)

    am_c_trms = np.sum((rfn * 32767) ** 2) / (32767**2 * n)
    print("am_c_trms = ", am_c_trms)

    am_c_tabs = np.sum(np.abs(rfn) * 32767) / (32767 * n)
    print("am_c_tabs = ", am_c_tabs)

    # assume that the isodelay point is at the peak
    am_c_sym = np.argmax(np.abs(rfn)) / n
    print("am_c_sym = ", am_c_sym)
